{
****************************************************************
*  Name    : SUART.BAS                                         *
*  Author  : David John Barker                                 *
*  Notice  : Copyright (c) 2006 Mecanique                      *
*          : All Rights Reserved                               *
*  Date    : 16 SEPT 2023                                      *
*  Version : 1.8b1 16 SEPT 2023                                *
*          : add parity option for formats 8N1, 8E1, or 8O1
*          :   SUART_PARITY : 0=none, 1=odd, 2=even
*          :  parity is ignored for rx
*  Version : 1.8 14 DEC 2022                                   *
*          : change ReadByte() and ReadByteTimeout() so they   *
*          :    have a ref to OneBT, otherwise you get a       *
*          :    'Symbol not previously defined (ONEBT)' error  *
*          :    if WriteByte() is not also used                *
*          : 1.7 5/18/2022                                     *
*          : add rx framing err check and FrameError flag      *
*          :   (sticky, once set must be cleared by user)      *
*          : add ReadByteTimeout() and RxTimeout flag          *
*          : various optimizations                             *
*          : replace rel braches (ie bra $-34) with labels     *
*          : add '#option SUART_DEBUG_PIN' for scope rx timing *
*          : 1.6 2/2/2021                                      *
*          : add '#define Z 2' and '#define C 0' to asm blocks *                  
*          : 1.5 17/03/2020                                    *
*          : add #option SUART_MODE and SUART_INIT_TX          *
*          : (see notes below)                                 *
*          : 1.4 19/08/2019                                    *
*          : add PORT/LAT/TRIS offset for devices with         *
*          :   different reg maps (K40, K42, Q10, etc)         *
*          :   now works with all but J94/J99 PORTK/PORTL      *
*          : add TX pin init to idle state                     *
*  Notes   : 1.3 28/05/2006 Changed "W" to "WREG" in ReadByte  *
*          : 1.2 Replaced 'rcall' with 'call'                  *
*          : 1.1 Release version                               *
****************************************************************
}
Module UART

Include "system.bas"

// parity options: 0=none (8N1), 1=odd (8O1), 2=even (8E1)
#if IsOption(SUART_PARITY) and not (SUART_PARITY in (0, 1, 2))
  #error SUART_PARITY, "Invalid parity option"
#endif
#option SUART_PARITY = 0        // default: none

// default UART Mode setting
// valid settings are 'umXXXX' values from the UART mode consts listed below
// set '#option SUART_MODE = 1' to emulate a normal hw uart (ie idle high)
#if isoption(SUART_MODE) and not(SUART_MODE in (1, 2, 4, 4+1, 4+2))
  #error "Invalid SUART_MODE setting"
#endif
#option SUART_MODE = 2     // default = umInverted
const SUART_DEFAULT_MODE = SUART_MODE

// this option controls TX pin intialization in SetTX()
//   true = pin is initialized to idle state
//   false = leave pin uninitialized (v1.3 default)
// note: for this to work properly you must either
//   a) change default '#option SUART_MODE', or
//   b) call SetMode() before calling SetTX()
// if this option is false then the TX pin state is left undefined
// after the call to SetTX()... LAT registers are not initialized at PON/RESET
#if isoption(SUART_INIT_TX) and not(SUART_INIT_TX in (true, false))
  #error "Invalid SUART_INIT_TX setting"
#endif
#option SUART_INIT_TX = true

// debug pin (for scope timing of RX input sampling point)
// pulses high when input RXD is sampled 
// example: #option SUART_DEBUG_PIN = PORTA.2
// to measure output timing just write a 'U' (0x55) and look at TX output
#if isoption(SUART_DEBUG_PIN)
  #warning "enabling SUART_DEBUG_PIN"
  dim SUART_DEBUG as SUART_DEBUG_PIN.SUART_DEBUG_PIN@
#endif
  
// if using the IDE Code Explorer to view this module, the system _clock 
// setting may not be visible here so the constants will display with different values
// than are actually used when compiling. 
// to view the actual settings, change the '#define _SYSCLOCK' manually to the 
// 'clock =' value in use (but be sure to change it back to '_clock' when done)
#define _SYSCLOCK = _clock      // insert your 'clock =' setting here
const SYSCLOCK = _SYSCLOCK
// make sure SYSCLOCK is set correctly to compile
#if (_SYSCLOCK <> _clock)
  #error "set '#define _SYSCLOCK = _clock' to build"
#endif
const FOSC = SYSCLOCK * 1000000             // clock, in Hz

Const 
   CycleCost = 20,                          // TX/RX instruction cycle count
   tcy300 = (2 * FOSC) / (4 * 300) / 2,     // 300 baud Tcy (3333.33us)
   tcy600 = (2 * FOSC) / (4 * 600) / 2,     // 600 baud Tcy (1666.66us)
   tcy1200 = (2 * FOSC) / (4 * 1200) / 2,   // 1200 baud Tcy (833.33us)
   tcy2400 = (2 * FOSC) / (4 * 2400) / 2,   // 2400 baud Tcy (416.66us)
   tcy4800 = (2 * FOSC) / (4 * 4800) / 2,   // 4800 baud Tcy (208.33us)
   tcy9600 = (2 * FOSC) / (4 * 9600) / 2,   // 9600 baud Tcy (104.16us)
   tcy19200 = (2 * FOSC) / (4 * 19200) / 2, // 19200 baud Tcy (52.08us)
   tcy38400 = (2 * FOSC) / (4 * 38400) / 2, // 38400 baud Tcy (26.04us) - requires clock >= 8MHZ
   tcy57600 = (2 * FOSC) / (4 * 57600) / 2  // 57600 baud Tcy (17.36us) - requires clock >= 16MHZ

// calculate baud rate constants
// these figures are one BitTime (BT) in microseconds (us)
// for example, 1/9600 baud gives 104.16 us
// taking into account the TX and RX cycle costs yields a figure of ~100 us
Public Const
   sbr300 As Word = (tcy300 - CycleCost) / (SYSCLOCK / 4),
   sbr600 As Word = (tcy600 - CycleCost) / (SYSCLOCK / 4),
   sbr1200 As Word = (tcy1200 - CycleCost) / (SYSCLOCK / 4),
   sbr2400 As Word = (tcy2400 - CycleCost) / (SYSCLOCK / 4),
   sbr4800 As Word = (tcy4800 - CycleCost) / (SYSCLOCK / 4),
   sbr9600 As Word = (tcy9600 - CycleCost) / (SYSCLOCK / 4),
   sbr19200 As Word = (tcy19200 - CycleCost) / (SYSCLOCK / 4),
   sbr38400 As Word = (tcy38400 - CycleCost) / (SYSCLOCK / 4),
   sbr57600 As Word = (tcy57600 - CycleCost) / (SYSCLOCK / 4)

// public UART modes...
Public Const
   umTrue = 1,                          // normal UART mode (idle high), bit 0
   umInverted = 2,                      // inverted UART (idle low), bit 1
   umOpen = 4,                          // open-drain, bit 2
   umOpenTrue = umOpen Or umTrue,
   umOpenInverted = umOpen Or umInverted


// status flag bits (in mode byte)
Public Const
   FERR_BIT = 7,
   RXTOUT_BIT = 6

// a pin structure...
Structure TPin
   PortAddr As Word                   // PORT addr (input), LAT addr (output)
   TrisAddr As Word                   // TRIS addr
   Pin      As Byte                   // pin bit mask
   PinMask  As Byte                   // inverted pin bit mask  
End Structure 

// private variables and aliases... 
Dim 
   FTX As TPin,                       // TX pin
   FRX As TPin,                       // RX pin
   FBaudrate As Word,                 // Baudrate (1 BT)
   FHalfBaudrate As Word,             // Half Baudrate (1/2 BT)
   FMode As Byte,                     // Mode (true, inverted, open-drain) and FERR status (bit 7)
   FIsTrue     As FMode.Booleans(0),  // true mode flag
   FIsInverted As FMode.Booleans(1),  // inverted mode flag
   FIsOD       As FMode.Booleans(2),  // open-drain mode flag
   FRxTimeout  As FMode.Booleans(6),  // read timeout status flag
   FFrameErr   As FMode.Booleans(7),  // framing error (FERR) status flag 
   FPacing As Word,                   // write byte pacing (us)
   FTerminator As Char                // read string terminator character

// public variables...
Public Dim
   ReadTerminator As FTerminator,     // read string terminator
   Pacing As FPacing,                 // write pacing (us) 
   FrameError As FFrameErr,           // FERR
   RxTimeout As FRxTimeout            // rx timeout

{
****************************************************************************
* Name    : LAT_OFFSET/TRIS_OFFSET                                         *
* Purpose : these macros load 'wresult' with the LAT and TRIS register     *
*         : offsets from PORT. they assume wresult is a 16-bit word        *
*         : located in the access bank (PROD is a good choice)             *
****************************************************************************
}
macro LAT_OFFSET(wresult)
    asm-
        movlw low(LATA-PORTA)
        movwf wresult
        movlw high(LATA-PORTA)
        movwf wresult+1
    end asm
end macro

macro TRIS_OFFSET(wresult)
    asm-
        movlw low(TRISA-PORTA)
        movwf wresult
        movlw high(TRISA-PORTA)
        movwf wresult+1
    end asm
end macro

{
****************************************************************************
* Name    : SetTX                                                          *
* Purpose : Set TX port and pin number                                     *
****************************************************************************
}
Public Sub SetTX(ByRef pPin As Bit)
   dim reg_offset as PROD      // 16-bit reg (must be in access bank)

   FTX.PortAddr = addressof(pPin)               // PORT addr
   TRIS_OFFSET(reg_offset)                      // get TRIS offset
   FTX.TrisAddr = FTX.PortAddr + reg_offset     // TRIS addr
   LAT_OFFSET(reg_offset)                       // get LAT offset
   FTX.PortAddr = FTX.PortAddr + reg_offset     // change PortAddr to LAT addr

   FTX.Pin = BitOf(pPin)               // get pin number (masked value)
   FTX.PinMask = Not FTX.Pin           // pin number inverted mask

 #if (SUART_INIT_TX)
   // set TX pin idle state... (added in v1.4)
   FSR0 = FTX.PortAddr
   If (FIsInverted) Then
      INDF0 = INDF0 And FTX.PinMask   
   Else
      INDF0 = INDF0 Or FTX.Pin
   EndIf
   FSR0 = FTX.TrisAddr
   If ((FMode And umOpen) = 0) Then
      INDF0 = INDF0 And FTX.PinMask     // set pin to output
   Else
      INDF0 = INDF0 Or FTX.Pin          // set pin to input (OD mode)
   EndIf
 #endif     //_SUART_INIT_TX
End Sub

{
****************************************************************************
* Name    : SetRX                                                          *
* Purpose : Set RX port and pin number                                     *
****************************************************************************
}
Public Sub SetRX(ByRef pPin As Bit)
   dim reg_offset as PROD      // 16-bit reg (must be in access bank)
   
   FRX.PortAddr = addressof(pPin)               // PORT addr
   TRIS_OFFSET(reg_offset)                      // get TRIS offset
   FRX.TrisAddr = FRX.PortAddr + reg_offset     // TRIS addr

   FRX.Pin = BitOf(pPin)               // get pin number (masked value)
   FRX.PinMask = Not FRX.Pin           // pin number inverted mask

   FSR0  = FRX.TrisAddr                // point FSR0 to TRIS
   INDF0 = INDF0 Or FRX.Pin            // set pin to input
End Sub

{
****************************************************************************
* Name    : SetBaudrate                                                    *
* Purpose : Sets the software UART baudrate                                *
*         : Pass a 'sbr' constant, as defined above. For example, sbr9600  *
****************************************************************************
}
Public Sub SetBaudrate(pBaudrate As Word)
   FBaudrate = pBaudrate
   FHalfBaudrate = FBaudrate / 2
End Sub

{
****************************************************************************
* Name    : SetMode                                                        *
* Purpose : Set the UART mode, either TRUE or INVERTED                     *
*         : Use a 'um' constant, as defined above (For example 'umTrue')   *       
*         : Module default is umInverted                                   *
****************************************************************************
}
Public Sub SetMode(pMode As Byte)
   FMode = pMode
End Sub

// parity functions
#if (SUART_PARITY > 0)
public const
    EVEN_PARITY = 0,
    ODD_PARITY = 1

// parity will be stored in msb of mode byte
public const
    PARITY_BIT = 7

// returns parity of b...
//   0 = even
//   1 = odd
public function parity(b as byte) as bit
    // breakdown the following into a simpler expression...
    // b = b xor ((b >> 4) or (b << 4))
    asm
        swapf b,W
    end asm
    b = b xor WREG

    // magic parity calculation
    b = b + $41
    b = b or $7c
    b = b + 2

    // check if (b >= 128)
    if (b.bits(7) = 1) then
        result = ODD_PARITY
    else
        result = EVEN_PARITY
    endif
end function

// 8-bit data mode...return parity so it can go it TX9
public function set_odd_parity(b as byte) as bit
    if (parity(b) = EVEN_PARITY) then
        result = 1
    else
        result = 0
    endif
end function

public function set_even_parity(b as byte) as bit
    if (parity(b) = ODD_PARITY) then
        result = 1
    else
        result = 0
    endif
end function
#endif      // SUART_PARITY

{
****************************************************************************
* Name    : OneBT (PRIVATE)                                                *
* Purpose : wait 1 bit time                                                *
****************************************************************************
}
// note: this routine has at least a 5Tcy overhead to load the rate and 
// call delayus, plus an addt'l 4 cycles if FBaudrate is outside bank 0
// to reduce the addtl delay, setting '#option SUART_USE_PROD_RATE = true'
// will use the PROD reg to hold a copy of FBaudrate, making it less 
// dependant on the location of FBaudrate
#option SUART_USE_PROD_RATE = true
#if (SUART_USE_PROD_RATE)
NoInline Sub OneBT()
   DelayUS(PROD)
End Sub  
#else
NoInline Sub OneBT()
   DelayUS(FBaudrate)
End Sub  
#endif

{
****************************************************************************
* Name    : HalfBT (PRIVATE)                                               *
* Purpose : wait 1/2 bit time                                              *
****************************************************************************
}
// note: this routine has at least a 5Tcy overhead to load the rate and 
// call delayus, plus an addt'l 4 cycles if FHalfBaudrate is outside bank 0
NoInline Sub HalfBT()
   DelayUS(FHalfBaudrate)
End Sub 

{
****************************************************************************
* Name    : ReadByte                                                       *
* Purpose : Read a byte from the software UART                             *
****************************************************************************
}
Public Function Access ReadByte() As Byte
   Dim Index, Pin, Mode As Byte
   Dim PORT_REG as INDF0    // PORT via FSR0
   
  // create asm symbols for SUART_DEBUG_PIN (if used)
  #if isoption(SUART_DEBUG_PIN)
    #option SUART_DEBUG_LAT = getLAT(SUART_DEBUG_PIN)   // create asm VARIABLE declarations
    dim _SUART_DEBUG_PORT as SUART_DEBUG_LAT
    const _SUART_DEBUG_PINNO = SUART_DEBUG_PIN@
  asm
  ifndef SUART_DEBUG_PORT
    #define SUART_DEBUG_PORT _SUART_DEBUG_PORT
    #define SUART_DEBUG_PINNO _SUART_DEBUG_PINNO
  endif
  end asm
  #endif    // isoption(SUART_DEBUG_PIN)

   // get module port, pin and mode into locals (access bank)
   FSR0 = FRX.PortAddr      // PORT address
   Pin = FRX.Pin            // get rx pin mask
   Mode = FMode
   Index = 8  

  #if (SUART_USE_PROD_RATE)
   PROD = FBaudrate
  #endif
     
ASM-
   ; check for existing STATUS Z and C bit definitions
   ifndef Z
    #define Z 2
   endif
   ifndef C
    #define C 0
   endif

   ; look For start bit...
rb_wait_start:
   movf    PORT_REG, W      ; <read pin>
   btfsc   Mode, 1          ; check umInverted...       
   comf    WREG             ; invert pin
   andwf   Pin, W           ; mask out other bits           
   btfss   STATUS, Z        ; pin is 0... found start       
   bra     rb_wait_start  
End ASM
   // reposition into the middle of the start bit time
   HalfBT()                   

ASM-
   ; recheck start bit...
 ifdef SUART_DEBUG_PORT
   bsf SUART_DEBUG_PORT, SUART_DEBUG_PINNO  ; DEBUG_PIN=1
 endif
   movf    PORT_REG, W      ; <read pin>
 ifdef SUART_DEBUG_PORT
   bcf SUART_DEBUG_PORT, SUART_DEBUG_PINNO  ; DEBUG_PIN=0
 endif
   btfsc   Mode, 1          ; check umInverted...       
   comf    WREG             ; invert pin
   andwf   Pin, W           ; mask out other bits           
   btfss   STATUS, Z        ; pin is 0... found start       
   bra rb_wait_start

   ; read data bits...
rb_read_bits:   
End ASM
   OneBT()          // ensure OneBT gets included
ASM-
 ifdef SUART_DEBUG_PORT
   bsf SUART_DEBUG_PORT, SUART_DEBUG_PINNO  ; DEBUG_PIN=1
 endif
   movf    PORT_REG, W      ; 1           | ** <read pin>
   bcf     STATUS, C        ; 1           | 1
   andwf   Pin, W           ; 1           | 2
   btfss   STATUS, Z        ; 1 (2)       | 3 (4)
   bsf     STATUS, C        ; 1           | 4
   rrcf    Result, F        ; 1           | 5
   bra     $ + 2            ; 2           | 7 <2-cycle nop>
   bra     $ + 2            ; 2           | 9 <2-cycle nop>
   bra     $ + 2            ; 2           | 11 <2-cycle nop>
 ifdef SUART_DEBUG_PORT
   bcf SUART_DEBUG_PORT, SUART_DEBUG_PINNO  ; DEBUG_PIN=0
 else         ; SUART_DEBUG already adds 2 cycles for debug pin bsf/bcf
   bra     $ + 2            ; 2           | 13 <2-cycle nop>
 endif
   decfsz  Index, F         ; 1 (2)       | 14 (15)
   bra     rb_read_bits     ; 2           | 16
END ASM

   // wait for parity bit and discard it
 #if (SUART_PARITY > 0)
   OneBT()
 #endif

ASM- 
   ; wait for stop bit and check for FERR
rb_wait_stop:
   incf    PRODL, F         ; fudge - increase stop bit wait time by 1us
   call    OneBT

   ; check stop bit
 ifdef SUART_DEBUG_PORT
   bsf SUART_DEBUG_PORT, SUART_DEBUG_PINNO  ; DEBUG_PIN=1
 endif
   movf    PORT_REG, W      ; <read pin>
 ifdef SUART_DEBUG_PORT
   bcf SUART_DEBUG_PORT, SUART_DEBUG_PINNO  ; DEBUG_PIN=0
 endif
   btfsc   Mode, 1          ; check umInverted...       
   comf    WREG             ; invert pin
   andwf   Pin, W           ; mask out other bits           
   btfss   STATUS, Z        ; pin is 0... framing error
   bra     rb_rx_done       ; pin is 1... stop bit ok  
rb_rx_ferr:
   banksel FMode
   bsf     FMode, FERR_BIT  ; set FERR framing error status flag
rb_rx_done:
End ASM

   If (FIsInverted) Then
      Result = Not Result
   EndIf  
   ClrWDT   
End Function

{
****************************************************************************
* Name    : ReadByteTimeout                                                *
* Purpose : Read a byte from the software UART with timeout                *
* Parameter: pTimeout timeout value, in msecs (0-65535)                    *
*          : If a timeout occurs while waiting for a start bit,            *
*          : RxTimeout is set true and the function returns 0              *
****************************************************************************
}
Public Function Access ReadByteTimeout(pTimeout as word) As Byte
   Dim Index, Pin, Mode As Byte
   Dim PORT_REG as INDF0    // PORT via FSR0
   Dim tout as longword
   
   // get port, pin and mode...
   Pin = FRX.Pin
   FSR0 = FRX.PortAddr      // PORT address
   Mode = FMode
   Index = 8  

   // convert pTimeout msecs to 'usecs' (VERY approximate)
   // we're going to cheat here and use shift to multiply by 1024 instead of 1000,
   // and since there's so much overhead in the timeout loop we'll scale back the
   // delay based on the clock
  #if (_clock > 32)
   tout = pTimeout << 10        // x 1024
  #elseif (_clock > 16)
   tout = pTimeout << 9         // x 512
  #elseif (_clock > 8)
   tout = pTimeout << 8         // x 256
  #else
   tout = pTimeout << 7         // x 128
  #endif
   
  #if (SUART_USE_PROD_RATE)
   PROD = FBaudrate
  #endif
     
   FRxTimeout = false           // reset timeout flag to start
   // look For start bit...
   while (true)
   ASM-
rbt_wait_start:
   movf    PORT_REG, W      ; <read pin>
   btfsc   Mode, 1                 
   comf    WREG             ; <- changed from "W"
   andwf   Pin, W                      
   bz      rbt_got_start    ; Z == 1 --> possible START bit              
   end asm
   if (tout = 0) then
      FRxTimeout = true     // set TIMEOUT flag
      result = 0
      exit
   endif
   tout = tout - 1
   end while
   // go on to read the byte...
asm
rbt_got_start:
end asm

   // position half-way into the bit time
   HalfBT()                   

ASM-
   ; check for existing STATUS Z and C bit definitions
   ifndef Z
    #define Z 2
   endif
   ifndef C
    #define C 0
   endif

   ; recheck start bit...
 ifdef SUART_DEBUG_PORT
   bsf SUART_DEBUG_PORT, SUART_DEBUG_PINNO  ; DEBUG_PIN=1
 endif
   movf    PORT_REG, W      ; <read pin>
 ifdef SUART_DEBUG_PORT
   bcf SUART_DEBUG_PORT, SUART_DEBUG_PINNO  ; DEBUG_PIN=0
 endif
   btfsc   Mode, 1          ; check umInverted...       
   comf    WREG             ; invert pin
   andwf   Pin, W           ; mask out other bits           
   btfss   STATUS, Z        ; pin is 0... found start       
   bra rbt_wait_start

   ; read data bits...
rbt_read_bits:   
End ASM
   OneBT()          // ensure OneBT gets included
ASM-
 ifdef SUART_DEBUG_PORT
   bsf SUART_DEBUG_PORT, SUART_DEBUG_PINNO  ; DEBUG_PIN=1
 endif
   movf    PORT_REG, W      ; 1           | ** <read pin>
   bcf     STATUS, C        ; 1           | 1
   andwf   Pin, W           ; 1           | 2
   btfss   STATUS, Z        ; 1 (2)       | 3 (4)
   bsf     STATUS, C        ; 1           | 4
   rrcf    Result, F        ; 1           | 5
   bra     $ + 2            ; 2           | 7 <2-cycle nop>
   bra     $ + 2            ; 2           | 9 <2-cycle nop>
   bra     $ + 2            ; 2           | 11 <2-cycle nop>
 ifdef SUART_DEBUG_PORT
   bcf SUART_DEBUG_PORT, SUART_DEBUG_PINNO  ; DEBUG_PIN=0
 else         ; SUART_DEBUG already adds 2 cycles for debug pin bsf/bcf
   bra     $ + 2            ; 2           | 13 <2-cycle nop>
 endif
   decfsz  Index, F         ; 1 (2)       | 14 (15)
   bra     rbt_read_bits    ; 2           | 16
END ASM

   // wait for parity bit and discard it
 #if (SUART_PARITY > 0)
   OneBT()
 #endif

ASM- 
   ; wait for stop bit and check for FERR
rbt_wait_stop:
   incf    PRODL, F         ; fudge - increase stop bit wait time by 1us
   call    OneBT

   ; check stop bit
 ifdef SUART_DEBUG_PORT
   bsf SUART_DEBUG_PORT, SUART_DEBUG_PINNO  ; DEBUG_PIN=1
 endif
   movf    PORT_REG, W      ; <read pin>
 ifdef SUART_DEBUG_PORT
   bcf SUART_DEBUG_PORT, SUART_DEBUG_PINNO  ; DEBUG_PIN=0
 endif
   btfsc   Mode, 1          ; check umInverted...       
   comf    WREG             ; invert pin
   andwf   Pin, W           ; mask out other bits           
   btfss   STATUS, Z        ; pin is 0... framing error
   bra     rbt_rx_done      ; pin is 1... stop bit ok  
rbt_rx_ferr:
   banksel FMode
   bsf     FMode, FERR_BIT  ; set FERR framing error status flag
rbt_rx_done:
End ASM

   If FIsInverted Then
      Result = Not Result
   EndIf  
   ClrWDT   
End Function

{
****************************************************************************
* Name    : WriteByte                                                      *
* Purpose : Write a byte to the software UART                              *
****************************************************************************
}
Public Sub Access WriteByte(pValue As Byte)  
   Dim Index, Mode, ModeOpen As Byte
   Dim Pin, PinMask As Byte
   
   // variable aliases
   Dim MakeHigh As Pin, MakeLow As PinMask
   Dim MakeInput As Pin, MakeOutput As PinMask
   Dim SaveTRIS As ModeOpen
   
   Dim LAT_REG as INDF0     // LAT via FSR0
   Dim TRIS_REG as INDF1    // TRIS via FSR1

   // get module port, pin and mode into locals (access bank)
   Mode = FMode

   ModeOpen = 0
   If (FIsInverted) Then
      ModeOpen = FTX.Pin
      pValue = Not pValue
   EndIf
   Pin = FTX.Pin              // pin mask
   PinMask = FTX.PinMask      // inverse mask
   FSR0 = FTX.PortAddr        // LAT address
   FSR1 = FTX.TrisAddr        // TRIS address

   // if enabled, compute parity and save it in msb of Mode
  #if (SUART_PARITY = 1)            // compute odd parity
   Mode.bits(PARITY_BIT) = set_odd_parity(pValue)
  #elseif (SUART_PARITY = 2)        // compute even parity
   Mode.bits(PARITY_BIT) = set_even_parity(pValue)
  #endif

  #if (SUART_USE_PROD_RATE)
   PROD = FBaudrate
  #endif
      
   Index = 9                  // 1 start, 8 data bits
ASM-
   ; check for existing STATUS Z and C bit definitions
   ifndef Z
    #define Z 2
   endif
   ifndef C
    #define C 0
   endif
   
   ; start bit...
   bsf    STATUS, C       ; assume start bit is high
   btfsc  Mode, 0         ; is it true mode
   bcf    STATUS, C       ; yes, so start bit is low
   bra    wb_write_bit    ; send the start bit
    
   ; main loop...
wb_write_bits:
   rrcf   pValue, F       ; 1           | 8

   ; set direction...  
wb_write_bit:
   movf   TRIS_REG, W     ; 1           | 9  <set TRIS>
   andwf  MakeOutput, W   ; 1           | 10   
   btfsc  STATUS, C       ; 1 (2)       | 11 (12)
   iorwf  MakeInput, W    ; 1           | 12 
   xorwf  ModeOpen, W     ; 1           | 13
   btfss  Mode, 2         ; 1 (2)       | 14 (15)
   andwf  MakeOutput, W   ; 1           | 15
   movwf  TRIS_REG        ; 1           | 16 <set TRIS>

   ; set bit value...
   movf   LAT_REG, W      ; 1           | 17 <write LAT>
   andwf  MakeLow, W      ; 1           | 18 
   btfsc  STATUS, C       ; 1 (2)       | 19 (20)
   iorwf  MakeHigh, W     ; 1           | 20 
   movwf  LAT_REG         ; 1           | ** <write LAT>
   call   OneBT           ; 2 + BT + 2  | 4
   decfsz Index, F        ; 1 (2)       | 5 (6)
   bra    wb_write_bits   ; 2           | 7
End ASM

#if (SUART_PARITY > 0)
Asm-
   ; parity bit...
   rlcf  Mode, W        ; get parity bit into C

   ; set direction...  
   movf   TRIS_REG, W     ; 1           | 9  <set TRIS>
   andwf  MakeOutput, W   ; 1           | 10   
   btfsc  STATUS, C       ; 1 (2)       | 11 (12)
   iorwf  MakeInput, W    ; 1           | 12 
   xorwf  ModeOpen, W     ; 1           | 13
   btfss  Mode, 2         ; 1 (2)       | 14 (15)
   andwf  MakeOutput, W   ; 1           | 15
   movwf  TRIS_REG        ; 1           | 16 <set TRIS>

   ; set bit value...
   movf   LAT_REG, W      ; 1           | 17 <write LAT>
   andwf  MakeLow, W      ; 1           | 18 
   btfsc  STATUS, C       ; 1 (2)       | 19 (20)
   iorwf  MakeHigh, W     ; 1           | 20 
   movwf  LAT_REG         ; 1           | ** <write LAT>
   call   OneBT           ; 2 + BT + 2  | 4
End ASM
#endif

Asm-
   ; stop bit, set direction...
   movf   TRIS_REG, W     ; 1           | 8 <set TRIS>
   iorwf  MakeInput, W    ; 1           | 9
   btfss  Mode, 2         ; 1 (2)       | 10 (11)
   andwf  MakeOutput, W   ; 1           | 11
   movwf  SaveTRIS        ; 1           | 12 

   ; stop bit, data...
   movf   LAT_REG, W      ; 1           | 13 <write LAT>
   andwf  MakeLow, W      ; 1           | 14
   btfsc  Mode, 0         ; 1 (2)       | 15 (16)
   iorwf  MakeHigh, W     ; 1           | 16
   bra    $ + 2           ; 2           | 18 <2-cycle nop>
   movff  SaveTRIS, TRIS_REG ; 2/3      | 20 <set TRIS>
   movwf  LAT_REG         ; 1           | ** <write LAT>
End ASM

   // wait stop bit time...
   // this also ensures that the subroutine is linked into 
   // program to allow above call from within the ASM block...
   OneBT()

   // wait any addtl tx pacing delay
   If (FPacing > 0) Then
      DelayUS(FPacing)
   EndIf 
   ClrWDT  
End Sub

{
****************************************************************************
* Name    : ReadBoolean                                                    *
* Purpose : Read a boolean from the software UART                          *
****************************************************************************
}
Public Function ReadBoolean() As Boolean
   Result = Boolean(ReadByte)
End Function 
{
****************************************************************************
* Name    : ReadWord                                                       *
* Purpose : Read a word from the software UART                             *
****************************************************************************
}
Public Function ReadWord() As Word
   Result.Bytes(0) = ReadByte
   Result.Bytes(1) = ReadByte
End Function  
{
****************************************************************************
* Name    : ReadLongWord                                                   *
* Purpose : Read a long word from the software UART                        *
****************************************************************************
}
Public Function ReadLongWord() As LongWord
   Result.Bytes(0) = ReadByte
   Result.Bytes(1) = ReadByte
   Result.Bytes(2) = ReadByte
   Result.Bytes(3) = ReadByte
End Function  
{
****************************************************************************
* Name    : ReadFloat                                                      *
* Purpose : Read a float from the software UART                            *
****************************************************************************
}
Public Function ReadFloat() As Float
   Result.Bytes(0) = ReadByte
   Result.Bytes(1) = ReadByte
   Result.Bytes(2) = ReadByte
   Result.Bytes(3) = ReadByte
End Function  

{
****************************************************************************
* Name    : WriteBoolean                                                   *
* Purpose : Write a boolean value to the software UART                     *
****************************************************************************
}
Public Sub WriteBoolean(pValue As Boolean)
   WriteByte(Byte(pValue))
End Sub
{
****************************************************************************
* Name    : WriteWord                                                      *
* Purpose : Write a word value to the software UART                        *
****************************************************************************
}
Public Sub WriteWord(pValue As Word)
   WriteByte(pValue.Bytes(0))
   WriteByte(pValue.Bytes(1))
End Sub
{
****************************************************************************
* Name    : WriteLongWord                                                  *
* Purpose : Write a word value to the software UART                        *
****************************************************************************
}
Public Sub WriteLongWord(pValue As LongWord) 
   WriteByte(pValue.Bytes(0))
   WriteByte(pValue.Bytes(1))
   WriteByte(pValue.Bytes(2))
   WriteByte(pValue.Bytes(3))
End Sub
{
****************************************************************************
* Name    : WriteFloat                                                     *
* Purpose : Write a floating point number to the software UART             *
****************************************************************************
}
Public Sub WriteFloat(pValue As Float) 
   WriteByte(pValue.Bytes(0))
   WriteByte(pValue.Bytes(1))
   WriteByte(pValue.Bytes(2))
   WriteByte(pValue.Bytes(3))
End Sub

{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a boolean from the software UART                          *
****************************************************************************
}
Sub ReadItem(ByRef pValue As Boolean)
   pValue = Boolean(ReadByte)
End Sub 
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a byte from the software UART                             *
****************************************************************************
}
Sub ReadItem(ByRef pValue As Byte)
   pValue = ReadByte
End Sub  
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a shortint from the software UART                         *
****************************************************************************
}
Sub ReadItem(ByRef pValue As ShortInt)
   pValue = ReadByte
End Sub  
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a word from the software UART                             *
****************************************************************************
}
Sub ReadItem(ByRef pValue As Word)
   pValue = ReadWord
End Sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read an integer from the software UART                         *
****************************************************************************
}
Sub ReadItem(ByRef pValue As Integer)
   pValue = ReadWord
End Sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a long word from the software UART                        *
****************************************************************************
}
Sub ReadItem(ByRef pValue As LongWord)
   pValue = ReadLongWord
End Sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a long integer from the software UART                     *
****************************************************************************
}
Sub ReadItem(ByRef pValue As LongInt)
   pValue = ReadLongWord
End Sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a floating point number from the software UART            *
****************************************************************************
}
Sub ReadItem(ByRef pValue As Float) 
   pValue = ReadFloat
End Sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Read a string from the software UART. Use ReadTerminator       *
*         : to specify the input string terminator character.              *
****************************************************************************
}
Sub ReadItem(ByRef pText As String)
   Dim Data, Index As Byte
   
   Index = 0
   Data = ReadByte
   While Data <> Byte(FTerminator)
      pText(Index) = Data
      Data = ReadByte
      Inc(Index)
   Wend
   pText(Index) = null
End Sub
{
****************************************************************************
* Name    : Read (COMPOUND)                                                *
* Purpose : Read an item from the software UART                            *
****************************************************************************
}
Public Compound Sub Read(ReadItem)

{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a boolean value to the software UART                     *
****************************************************************************
}
Sub WriteItem(pValue As Boolean)
   WriteByte(Byte(pValue))
End Sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a byte value to the software UART                        *
****************************************************************************
}
Sub WriteItem(pValue As WREG)
   WriteByte(pValue)
End Sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a short int value to the software UART                   *
****************************************************************************
}
Sub WriteItem(pValue As ShortInt)
   WriteByte(pValue)
End Sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a word value to the software UART                        *
****************************************************************************
}
Sub WriteItem(pValue As Word) 
   WriteWord(pValue)
End Sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write an integer value to the software UART                    *
****************************************************************************
}
Sub WriteItem(pValue As Integer) 
   WriteWord(pValue)
End Sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a long word to the software UART                         *
****************************************************************************
}
Sub WriteItem(pValue As LongWord) 
   WriteLongWord(pValue)
End Sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a long integer to the software UART                      *
****************************************************************************
}
Sub WriteItem(pValue As LongInt) 
   WriteLongWord(pValue)
End Sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a floating point number to the software UART             *
****************************************************************************
}
Sub WriteItem(pValue As Float) 
   WriteByte(pValue.Bytes(0))
   WriteByte(pValue.Bytes(1))
   WriteByte(pValue.Bytes(2))
   WriteByte(pValue.Bytes(3))
End Sub
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Write a string value to the software UART                      *
****************************************************************************
}
Sub WriteItem(pText As String)
   Dim Index, Data As Byte  
   Index = 0
   Data = pText(0)
   While Data <> 0
      WriteByte(Data)
      Inc(Index)
      Data = pText(index)
   Wend
End Sub
{
****************************************************************************
* Name    : Write (COMPOUND)                                               *
* Purpose : Write an item to the software UART                             *
****************************************************************************
}
Public Compound Sub Write(WriteItem)

//
// initialise module...
//
FTerminator = #13
FPacing = 0
FMode = SUART_DEFAULT_MODE

#if isoption(SUART_DEBUG_PIN)
low(SUART_DEBUG)
#endif

End Module

